//bmp.c

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include "types.h"
#include "bmp.h"
#include "bmp_server.h"

#pragma pack(1)

BITMAPFILEHEADER bhfi;
BITMAPINFOHEADER bhfo;
BITMAPINFOHEADERr bhfor;
BMP_PAR bmppar;
U8 BMP_DUMP[_BMP_DUMP_SIZE_];
U8 BMP_REDRAW_WORK;
U8 BMP_FILE;


/*
*************************************
* Анализ параметров BMP файла       *
*************************************
*/
int BMP_FileAnalise(char *fname)
{
    FILE * file;
    printf("BMP filename:  %s\n", fname);
    file = fopen(fname,"rb");
    if(file < 0) {
        printf("Open %s  error!\n",fname);
        return -1;
    }
    fseek(file, 0, 0);
    fread(&bhfi, sizeof(BITMAPFILEHEADER), 1, file);
    fseek(file, _bcSize, 0);
    fread(&bhfor.biSize, sizeof(DWORD), 1, file);    
    fseek(file, _bcSize, 0);
    fread(&bhfor.biSize, bhfor.biSize, 1, file);
    //чтение файла в буфер
    fseek(file,0,0);
    fread(&BMP_DUMP[0],1,bhfi.bfSize,file);
    //закрытие файла
    fclose(file);
    BMP_FILE = 0xFF;
    //анализ заголовка INFO
    printf("Size Struct info %u dec   %x hex\n",bhfor.biSize, bhfor.biSize);
    switch(bhfor.biSize)
    {
        case  12:
            printf("Version \"Core\"\n");
            (*(LONG*)(&bhfo.biWidth)) = (LONG)(*(WORD*)(&bhfor.data[0])); 
            (*(LONG*)(&bhfo.biHeight)) = (LONG)(*(WORD*)(&bhfor.data[2])); 
            (*(WORD*)(&bhfo.biBitCount)) = (*(WORD*)(&bhfor.data[6])); 
            break;
        case 40:
            printf("Version \"3\"\n");
            (*(LONG*)(&bhfo.biWidth)) = (*(LONG*)(&bhfor.data[0])); 
            (*(LONG*)(&bhfo.biHeight)) = (*(LONG*)(&bhfor.data[4])); 
            (*(WORD*)(&bhfo.biBitCount)) = (*(WORD*)(&bhfor.data[10])); 
            break;
        case 108:
            printf("Version \"4\"\n");
            (*(LONG*)(&bhfo.biWidth)) = (*(LONG*)(&bhfor.data[0])); 
            (*(LONG*)(&bhfo.biHeight)) = (*(LONG*)(&bhfor.data[4])); 
            (*(WORD*)(&bhfo.biBitCount)) = (*(WORD*)(&bhfor.data[10])); 
            break;
        case 124:
            printf("Version \"5\"\n");
            (*(LONG*)(&bhfo.biWidth)) = (*(LONG*)(&bhfor.data[0])); 
            (*(LONG*)(&bhfo.biHeight)) = (*(LONG*)(&bhfor.data[4])); 
            (*(WORD*)(&bhfo.biBitCount)) = (*(WORD*)(&bhfor.data[10])); 
            break;
        default:
            break;
    }
    //
    bmppar.step_x = bhfo.biBitCount / 8;
    printf("--- step x %u\r\n", bmppar.step_x);
    //bmppar.step_y =(LONG)((bhfo.biWidth + 1) * bmppar.step_x);
    bmppar.step_y =(LONG)((bhfo.biWidth) * bmppar.step_x);
    printf("--- step y %u\r\n", bmppar.step_y);
    //
    printf("---------------------------------------------------------------------------------\n");
    printf("Type         : %x\n", bhfi.bfType);
    printf("Size          : %u bytes\n", bhfi.bfSize);
    printf("Start addr: %x hex\n", bhfi.bfOffBits);
    printf("---------------------------------------------------------------------------------\n");
    printf("Widht :  %u\n", bhfo.biWidth);
    printf("Haight : %u\n", bhfo.biHeight);
    printf("Bit in Pixel(color) %u\n", bhfo.biBitCount);
    //
    return 1;
}


int BMP_SetPixel(FILE * file, LONG x, LONG y, U8 R, U8 G, U8 B)
{
    DWORD addr;
    WORD step_x = bmppar.step_x;
    LONG step_y = bmppar.step_y;
    RGBQUAD rgb;
    rgb.rgbRed = R;
    rgb.rgbGreen = G;
    rgb.rgbBlue = B;
    //проверка корректности координат
    if((x > bhfo.biWidth) ||
        (y > bhfo.biHeight))
            return -1;
    //
    addr = (step_x * x) + (step_y * y) + bhfi.bfOffBits;
    if(BMP_REDRAW_WORK == 0)
    {
        fseek(file, addr,0);
        fwrite(&rgb, step_x, 1, file);
    }
    else
        memcpy(&BMP_DUMP[addr], &rgb, step_x);
    //
    return 0;
}

int BMP_GetPixel(FILE * file, LONG x, LONG y, U8 * R, U8 * G, U8 * B)
{
	DWORD addr;
    	WORD step_x = bmppar.step_x;
    	LONG step_y = bmppar.step_y;
    	RGBQUAD rgb;
	//проверка корректности координат
    	if((x > bhfo.biWidth) ||
        	(y > bhfo.biHeight))
		return -1;
	addr = (step_x * x) + (step_y * y) + bhfi.bfOffBits;
	if(BMP_REDRAW_WORK == 0)
    	{
        	fseek(file, addr,0);
        	//fwrite(&rgb, step_x, 1, file);
		fread(&rgb, step_x, 1, file);
    	}
    	else
         	//memcpy(&BMP_DUMP[addr], &rgb, step_x);
		memcpy(&rgb, &BMP_DUMP[addr], step_x);
	//
	*R = rgb.rgbRed;
	*G = rgb.rgbGreen;
	*B = rgb.rgbBlue; 
	//
	return 0;
}

inline int iBMP_SetPixel(FILE * file, LONG x, LONG y, U8 R, U8 G, U8 B)
{
    DWORD addr;
    WORD step_x = bmppar.step_x;
    LONG step_y = bmppar.step_y;
    RGBQUAD rgb;
    rgb.rgbRed = R;
    rgb.rgbGreen = G;
    rgb.rgbBlue = B;
    //проверка корректности координат
    if((x > bhfo.biWidth) ||
        (y > bhfo.biHeight))
            return -1;
    //
    addr = (step_x * x) + (step_y * y) + bhfi.bfOffBits;
    if(BMP_REDRAW_WORK == 0)
    {
        fseek(file, addr,0);
        fwrite(&rgb, step_x, 1, file); 
    }
    else
        memcpy(&BMP_DUMP[addr], &rgb, step_x);
    //
    return 0;
}

inline void BMP_hline(FILE * file, U32 x, U32 y, U32 len, U8 R, U8 G, U8 B)
{
    DWORD addr,addr1;
    WORD step_x = bmppar.step_x;
    LONG step_y = bmppar.step_y;
    RGBQUAD rgb;
    rgb.rgbRed = R;
    rgb.rgbGreen = G;
    rgb.rgbBlue = B;
    addr1 = (step_y * y) + bhfi.bfOffBits;
	while(len--)
	{
        addr = addr1 + (step_x * x);
        if(BMP_REDRAW_WORK == 0)
        {
            fseek(file, addr,0);
            fwrite(&rgb, step_x, 1, file); 
        }
        else
            memcpy(&BMP_DUMP[addr], &rgb, step_x);
	    x++;
	}    
}


inline void BMP_vline(FILE * file, U32 x, U32 y, U32 len, U8 R, U8 G, U8 B)
{
    DWORD addr,addr1;
    WORD step_x = bmppar.step_x;
    LONG step_y = bmppar.step_y;
    RGBQUAD rgb;
    rgb.rgbRed = R;
    rgb.rgbGreen = G;
    rgb.rgbBlue = B;
    addr1 = (step_x * x) + bhfi.bfOffBits;
    while(len--)
	{
        addr = addr1 + (step_y * y);
        if(BMP_REDRAW_WORK == 0)
        {
            fseek(file, addr,0);
            fwrite(&rgb, step_x, 1, file); 
        }
        else
            memcpy(&BMP_DUMP[addr], &rgb, step_x);
	    y++;
	}

}

/*******
 *  line *
 *******/
 static void do_line(FILE * file, int x1, int y1, int x2, int y2, BYTE R, BYTE G, BYTE B)
{
   int dx = x2-x1;
   int dy = y2-y1;
   int i1, i2;
   int x, y;
   int dd;

   /* worker macro */
   #define DO_LINE(pri_sign, pri_c, pri_cond, sec_sign, sec_c, sec_cond)     \
   {                                                                         \
      if (d##pri_c == 0) {                                                   \
    iBMP_SetPixel(file, x1, y1, R, G, B);                     \
	 return;                                                             \
      }                                                                      \
									     \
      i1 = 2 * d##sec_c;                                                     \
      dd = i1 - (sec_sign (pri_sign d##pri_c));                              \
      i2 = dd - (sec_sign (pri_sign d##pri_c));                              \
									     \
      x = x1;                                                                \
      y = y1;                                                                \
									     \
      while (pri_c pri_cond pri_c##2) {                                      \
        iBMP_SetPixel(file, x, y, R, G, B);                                                 \
									     \
	 if (dd sec_cond 0) {                                                \
	    sec_c sec_sign##= 1;                                             \
	    dd += i2;                                                        \
	 }                                                                   \
	 else                                                                \
	    dd += i1;                                                        \
									     \
	 pri_c pri_sign##= 1;                                                \
      }                                                                      \
   }

   if (dx >= 0) {
      if (dy >= 0) {
	 if (dx >= dy) {
	    /* (x1 <= x2) && (y1 <= y2) && (dx >= dy) */
	    DO_LINE(+, x, <=, +, y, >=);
	 }
	 else {
	    /* (x1 <= x2) && (y1 <= y2) && (dx < dy) */
	    DO_LINE(+, y, <=, +, x, >=);
	 }
      }
      else {
	 if (dx >= -dy) {
	    /* (x1 <= x2) && (y1 > y2) && (dx >= dy) */
	    DO_LINE(+, x, <=, -, y, <=);
	 }
	 else {
	    /* (x1 <= x2) && (y1 > y2) && (dx < dy) */
	    DO_LINE(-, y, >=, +, x, >=);
	 }
      }
   }
   else {
      if (dy >= 0) {
	 if (-dx >= dy) {
	    /* (x1 > x2) && (y1 <= y2) && (dx >= dy) */
	    DO_LINE(-, x, >=, +, y, >=);
	 }
	 else {
	    /* (x1 > x2) && (y1 <= y2) && (dx < dy) */
	    DO_LINE(+, y, <=, -, x, <=);
	 }
      }
      else {
	 if (-dx >= -dy) {
	    /* (x1 > x2) && (y1 > y2) && (dx >= dy) */
	    DO_LINE(-, x, >=, -, y, <=);
	 }
	 else {
	    /* (x1 > x2) && (y1 > y2) && (dx < dy) */
	    DO_LINE(-, y, >=, -, x, <=);
	 }
      }
   }
}

 
 
void BMP_Line(FILE * file, U32  x1, U32 y1, U32  x2, U32 y2, U8 R, U8 G, U8 B)
{
     int sx, sy, dx, dy, t;
   
      sx = x1;
      sy = y1;
      dx = x2;
      dy = y2;

      if (sx > dx) {
        t = sx;
        sx = dx;
        dx = t;
      }

      if (sy > dy) {
        t = sy;
        sy = dy;
        dy = t;
      }


    if (x1 == x2) {
      if (y2>y1)
            BMP_vline(file, x1, y1, y2-y1, R, G, B);
      else
            BMP_vline(file, x1, y2, y1-y2, R, G, B);
      return;
   }

   if (y1 == y2) {
      if (x2>x1)
         BMP_hline(file, x1, y1, x2-x1, R, G, B);
      else
        BMP_hline(file, x2, y1, x1-x2, R, G, B);
      return;
   }

   do_line(file,x1, y1, x2, y2, R, G, B);
}


/********
 * circle *
 ********/
static void do_circle(FILE * file, int x, int y, int radius, BYTE R, BYTE G, BYTE B)
{
   int cx = 0;
   int cy = radius;
   int df = 1 - radius; 
   int d_e = 3;
   int d_se = -2 * radius + 5;

   do {
       iBMP_SetPixel(file, x+cx, y+cy, R, G, B);

      if (cx) 
        iBMP_SetPixel(file, x-cx, y+cy, R, G, B);

      if (cy) 
        iBMP_SetPixel(file, x+cx, y-cy, R, G, B);

      if ((cx) && (cy)) 
        iBMP_SetPixel(file, x-cx, y-cy, R, G, B);

      if (cx != cy) {
        iBMP_SetPixel(file, x+cy, y+cx, R, G, B);

	 if (cx) 
        iBMP_SetPixel(file, x+cy, y-cx, R, G, B);

	 if (cy) 
        iBMP_SetPixel(file, x-cy, y+cx, R, G, B);

	 if (cx && cy) 
        iBMP_SetPixel(file, x-cy, y-cx, R, G, B);
      }

      if (df < 0)  {
        df += d_e;
        d_e += 2;
        d_se += 2;
      }
      else { 
        df += d_se;
        d_e += 2;
        d_se += 4;
        cy--;
      } 

      cx++; 

   } while (cx <= cy);
}

void BMP_Circle(FILE * file, U32 x, U32 y, U32 radius, U8 R, U8 G, U8 B)
{
   int sx, sy, dx, dy;

    sx = x-radius-1;
    sy = y-radius-1;
    dx = x+radius+1;
    dy = y+radius+1;

    do_circle(file, x, y, radius, R, G, B);
}

void BMP_Circlefill(FILE * file, U32 x, U32 y, U32 radius, U8 R, U8 G, U8 B)
{
   int cx = 0;
   int cy = radius;
   int df = 1 - radius; 
   int d_e = 3;
   int d_se = -2 * radius + 5;
   int sx, sy, dx, dy;


    sx = x-radius-1;
    sy = y-radius-1;
    dx = x+radius+1;
    dy = y+radius+1;

    do
    {
        BMP_hline(file, x-cy, y-cx, cy*2, R, G, B);
        if (cx)
            BMP_hline(file, x-cy, y+cx, cy*2, R, G, B);
    	if (df < 0)
        {
            df += d_e;
            d_e += 2;
            d_se += 2;
        }
        else 
        { 
            if (cx != cy)
            {
                BMP_hline(file, x-cx, y-cy, cx*2, R, G, B);
                if (cy)
                      BMP_hline(file, x-cx, y+cy, cx*2, R, G, B);
            }

            df += d_se;
            d_e += 2;
            d_se += 4;
            cy--;
        } 
        cx++; 
    } while (cx <= cy);
}

void BMP_Circlefill_fuzzy(FILE * file, U32 x, U32 y, U32 radius, BYTE R, BYTE G, BYTE B, int f)
{
   int cx = 0;
   int cy = radius;
   int df = 1 - radius; 
   int d_e = 3;
   int d_se = -2 * radius + 5;
   int sx, sy, dx, dy;
    

    sx = x-radius-1;
    sy = y-radius-1;
    dx = x+radius+1;
    dy = y+radius+1;

     do
    {
        //gfx_hline(x-cy-f, y-cx, (cy+f)*2, color);
        BMP_hline(file, x-cy-f, y-cx, (cy+f)*2, R, G, B);
        if (cx)
            //gfx_hline(x-cy-f, y+cx, (cy+f)*2, color);
            BMP_hline(file, x-cy-f, y+cx, (cy+f)*2, R, G, B);
    	if (df < 0)
        {
            df += d_e;
            d_e += 2;
            d_se += 2;
        }
        else 
        { 
            if (cx != cy)
            {
                //gfx_hline(x-cx-f, y-cy, (cx+f)*2, color);
                BMP_hline(file, x-cx-f, y-cy, (cx+f)*2, R, G, B);
            if (cy)
	    	    //gfx_hline(x-cx-f, y+cy, (cx+f)*2, color);
                BMP_hline(file, x-cx-f, y+cy, (cx +f)*2, R, G, B);
            }

            df += d_se;
            d_e += 2;
            d_se += 4;
            cy--;
        } 
        if(f==1)
            f = -1;
        else
            f = 1;        
        cx++; 
    } while (cx <= cy);
}



/****************
 * RECTANGLE *
 ****************/

void BMP_rect(FILE * file, U32 x1, U32 y1, U32 x2, U32 y2, BYTE R, BYTE G, BYTE B)
{
    int w, h, t;

    
    if (x1>x2)
    {
        t = x1;
        x1 = x2;
        x2 = t;
    }
    if (y1>y2)
    {
        t = y1;
        y1 = y2;
        y2 = t;
    }

    w = x2 - x1 + 1;
    h = y2 - y1 + 1;
    
   //gfx_hline(x1, y1, w, color);
   BMP_hline(file, x1, y1, w, R, G, B);
   if (h>1)
	//gfx_hline(x1, y1 + h -1, w, color); 
    BMP_hline(file, x1, y1+h-1, w, R, G, B);
   //gfx_vline(x1, y1, h, color);
   BMP_vline(file, x1, y1, h, R, G, B);
   if (w>1)
	//gfx_vline(x1 + w -1, y1, h, color);
    BMP_vline(file, x1+w-1, y1, h, R, G, B);
}

void BMP_rect_filled(FILE * file, U32 x1, U32 y1, U32 x2, U32 y2, BYTE R, BYTE G, BYTE B)
{
    int w, h, t;
//    if (drop_next_frame)
//	return;

    if (x1>x2)
    {
        t = x1;
        x1 = x2;
        x2 = t;
    }
    if (y1>y2)
    {
        t = y1;
        y1 = y2;
        y2 = t;
    }

    w = x2 - x1 + 1;
    h = y2 - y1 + 1;
    
    while(h--)    
    {
        //gfx_hline(x1, y1, w, color);
        BMP_hline(file, x1, y1, w, R, G, B);
        y1++ ;
    }
}


/**************
 * TRIANGLE *
 **************/
 void BMP_triangle(FILE * file, U32 x1, U32  y1, U32 x2, U32 y2, U32 x3, U32 y3, BYTE R, BYTE G, BYTE B)
{
   //gfx_line(x1, y1, x2, y2, color);
   BMP_Line(file, x1, y1, x2, y2, R, G, B);
   //gfx_line(x2, y2, x3, y3, color);
   BMP_Line(file, x2, y2, x3, y3, R, G, B);
   //gfx_line(x1, y1, x3, y3, color);
   BMP_Line(file, x1, y1, x3, y3, R, G, B);
}



/*
 ****************************
 * поток перезаписи файла *
 ****************************
 */
 void pthBMP_Redraw(void)
 {
    printf("starting loop redraw.....\r\n");
    while(BMP_REDRAW_WORK > 0)
    {
        if((BMP_FILE > 0) && (dat_bmp_server.ffile > 0))
        {
            fseek(dat_bmp_server.ffile , 0,0);
            fwrite(&BMP_DUMP[0], 1, bhfi.bfSize, dat_bmp_server.ffile );
        }
	else
		printf("***  FILE ERROR: %i  ***\r\n", dat_bmp_server.ffile);
        //
        usleep(_interval_redraw);
    }
    printf("stop loop redraw.....\r\n");
 }
