//bmp_server.c

#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "types.h"
#include "socket.h"
#include "bmp_server.h"
#include "bmp.h"
#include "main.h"

#pragma pack(1)

DAT_BMP_SERVER dat_bmp_server;


pthread_t thread1, thread2;
pthread_t thBMP;
pthread_mutex_t mt;

/*
 inline блок клиента
*************************************************************************/
//бинарный формат
inline int iBMP_CLIENT_bin (BYTE * dumpi, BYTE * dumpo)
{
    int sizeout = 0;
    //U32 u32p;
    //
    //
    U32     SzPkt;
    U16     Cmd;
    U32     X1;
    U32     Y1;
    U32     X2;
    U32     Y2;
    BYTE    R;
    BYTE    G;
    BYTE    B;
    //обработка данных
    memcpy(&SzPkt, dumpi + _size_in_addr, sizeof(U32));
    memcpy(&Cmd, dumpi + _incmd_addr, sizeof(U16));
    switch(Cmd)
    {
        case _bmp_cmd_nop:  //пустая команда
            sizeout = 8;
            (*(U8*)(dumpo + _id_out_bin_addr)) = _id_bin;
            (*(U32*)(dumpo + _size_out_addr)) = sizeout;
            (*(U16*)(dumpo + _outcmd_addr)) = _bmp_cmd_nop;
            (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_ok;
            break;
        case _bmp_cmd_openfile: //открыть файл для работы
            //
            dat_bmp_server.ffile = fopen((char*)(dumpi + _incmd_filename),"rb");
            if(dat_bmp_server.ffile > 0)
            {
                fclose(dat_bmp_server.ffile);
                BMP_FileAnalise((char*)(dumpi + _incmd_filename));
                dat_bmp_server.ffile = fopen((char*)(dumpi + _incmd_filename),"r+b");
                (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_ok;
            }
            else
                (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_err;
            //
            (*(U8*)(dumpo + _id_out_bin_addr)) = _id_bin;
            (*(U32*)(dumpo + _size_out_addr)) = sizeout;
            (*(U16*)(dumpo + _outcmd_addr)) = _bmp_cmd_nop;
            
            break;
        case _bmp_cmd_setpixel: //установить пиксел
            sizeout = 8;
            if(dat_bmp_server.ffile > 0)
            {
                memcpy(&X1, dumpi + _addr_setpixel_X1, sizeof(U32));
                memcpy(&Y1, dumpi + _addr_setpixel_Y1, sizeof(U32));
                R = (*(U8*)(dumpi + _addr_setpixel_R));
                G = (*(U8*)(dumpi + _addr_setpixel_G));
                B = (*(U8*)(dumpi + _addr_setpixel_B));
                BMP_SetPixel((FILE*)dat_bmp_server.ffile, X1, Y1, R, G, B);
                BMP_SetPixel((FILE*)dat_bmp_server.ffile, X1, Y1, R, G, B);
                (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_ok;
            }
            else
                (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_notposible;
            (*(U8*)(dumpo + _id_out_bin_addr)) = _id_bin;
            (*(U32*)(dumpo + _size_out_addr)) = sizeout;
             memcpy(dumpo + _outcmd_addr, &Cmd, sizeof(U16));
            break;
        case _bmp_cmd_line: //нарисовать линию
            sizeout = 8;
            if(dat_bmp_server.ffile > 0)
            {
                memcpy(&X1, dumpi + _addr_line_X1, sizeof(U32));
                memcpy(&Y1, dumpi + _addr_line_Y1, sizeof(U32));
                memcpy(&X2, dumpi + _addr_line_X2, sizeof(U32));
                memcpy(&Y2, dumpi + _addr_line_Y2, sizeof(U32));
                R = (*(U8*)(dumpi + _addr_line_R));
                G = (*(U8*)(dumpi + _addr_line_G));
                B = (*(U8*)(dumpi + _addr_line_B));
                BMP_Line((FILE*)dat_bmp_server.ffile, X1, Y1, X2, Y2, R, G, B);
                BMP_Line((FILE*)dat_bmp_server.ffile, X1, Y1, X2, Y2, R, G, B);
                (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_ok;
            }
            else
                (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_notposible;
            (*(U8*)(dumpo + _id_out_bin_addr)) = _id_bin;
            (*(U32*)(dumpo + _size_out_addr)) = sizeout;
             memcpy(dumpo + _outcmd_addr, &Cmd, sizeof(U16));
            break;
        default:
            sizeout = 8;
            (*(U8*)(dumpo + _id_out_bin_addr)) = _id_bin;
            memcpy(dumpo + _size_out_addr , &sizeout, sizeof(U32));
            memcpy(dumpo + _outcmd_addr, &Cmd, sizeof(U16));
            (*(U8*)(dumpo + _outcmd_stat_addr)) = _bmp_cmd_stat_none;
            break;
    }
    return sizeout;
}



//строчный формат
inline int iBMP_CLIENT_string (BYTE * dumpi, BYTE * dumpo)
{
    int sizeout = 0;
    int count = 0;
    char par[32][32];
    int nump = 0;
    int countp = 0;
    U8 flag = 0;
    //
    U32 x1;
    U32 y1;
    U32 x2;
    U32 y2;
    U32 x3;
    U32 y3;
    U32 radius;
    BYTE r;
    BYTE g;
    BYTE b;
    int fuzzy;
    //
    while(count < 32)
    {
        memset(&par[count][0],0,32);
        count++;
    }
    //разложить строку по отдельным параметрам
    count = 0;
    while(dumpi[count] != 0)
    {
        if(dumpi[count] == 0x20)
        {
            if(!(flag > 0))
            {
                nump++;
                flag = 0xFF;
                countp = 0;
            }
        }
        else
        {
            memcpy(&par[nump][countp],(char*)(dumpi + count),1);
            flag = 0;
            countp++;
        }
        count++;
        if(count >= 256)
            break;
    }
    //интерпретатор
    if((strcmp(&par[0][0], "fileopen") == 0))
    {
        if(dat_bmp_server.ffile > 0)
            sizeout = sprintf((char*)(dumpo), "Opened\r\n");
        else
        {
            dat_bmp_server.ffile = fopen((char*)(&par[1][0]),"rb");
            if(dat_bmp_server.ffile > 0)
            {
                fclose(dat_bmp_server.ffile);
                BMP_FileAnalise((char*)(&par[1][0]));
                dat_bmp_server.ffile = fopen((char*)(&par[1][0]),"r+b");
                sizeout = sprintf((char*)(dumpo), "Ok\r\n");
            }
            else
                sizeout = sprintf((char*)(dumpo), "Error open file\r\n");
        }
    }
    else if((strcmp(&par[0][0], "fileclose") == 0))
    {
        fclose(dat_bmp_server.ffile);
        sizeout = sprintf((char*)(dumpo), "Ok\r\n");
    }
    else if((strcmp(&par[0][0], "setpixel") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            r = (BYTE)(atoi(&par[3][0]));
            g = (BYTE)(atoi(&par[4][0]));
            b = (BYTE)(atoi(&par[5][0]));
            if((BMP_SetPixel((FILE*)dat_bmp_server.ffile, x1, y1, r, g, b) < 0))
                    sizeout = sprintf((char*)(dumpo), "error set pixel\r\n");
            else
            {
                    BMP_SetPixel((FILE*)dat_bmp_server.ffile, x1, y1, r, g, b);
                    sizeout = sprintf((char*)(dumpo), "Ok\r\n");
            }
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "circle") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            radius = (U32)(atoi(&par[3][0]));
            r = (BYTE)(atoi(&par[4][0]));
            g = (BYTE)(atoi(&par[5][0]));
            b = (BYTE)(atoi(&par[6][0]));
            BMP_Circle((FILE*)dat_bmp_server.ffile, x1, y1, radius, r, g, b);
            BMP_Circle((FILE*)dat_bmp_server.ffile, x1, y1, radius, r, g, b);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "circlefill") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            radius = (U32)(atoi(&par[3][0]));
            r = (BYTE)(atoi(&par[4][0]));
            g = (BYTE)(atoi(&par[5][0]));
            b = (BYTE)(atoi(&par[6][0]));
            BMP_Circlefill((FILE*)dat_bmp_server.ffile, x1, y1, radius, r, g, b);
            BMP_Circlefill((FILE*)dat_bmp_server.ffile, x1, y1, radius, r, g, b);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "circlefill_fuzzy") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            radius = (U32)(atoi(&par[3][0]));
            r = (BYTE)(atoi(&par[4][0]));
            g = (BYTE)(atoi(&par[5][0]));
            b = (BYTE)(atoi(&par[6][0]));
            fuzzy = (int)(atoi(&par[7][0]));
            BMP_Circlefill_fuzzy((FILE*)dat_bmp_server.ffile, x1, y1, radius, r, g, b, fuzzy);
            BMP_Circlefill_fuzzy((FILE*)dat_bmp_server.ffile, x1, y1, radius, r, g, b, fuzzy);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "rect") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            x2 = (U32)(atoi(&par[3][0]));
            y2 = (U32)(atoi(&par[4][0]));
            r = (BYTE)(atoi(&par[5][0]));
            g = (BYTE)(atoi(&par[6][0]));
            b = (BYTE)(atoi(&par[7][0]));
            BMP_rect((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, r, g, b);
            BMP_rect((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, r, g, b);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "rect_filled") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            x2 = (U32)(atoi(&par[3][0]));
            y2 = (U32)(atoi(&par[4][0]));
            r = (BYTE)(atoi(&par[5][0]));
            g = (BYTE)(atoi(&par[6][0]));
            b = (BYTE)(atoi(&par[7][0]));
            BMP_rect_filled((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, r, g, b);
            BMP_rect_filled((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, r, g, b);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "triangle") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            x2 = (U32)(atoi(&par[3][0]));
            y2 = (U32)(atoi(&par[4][0]));
            x3 = (U32)(atoi(&par[5][0]));
            y3 = (U32)(atoi(&par[6][0]));
            r = (BYTE)(atoi(&par[7][0]));
            g = (BYTE)(atoi(&par[8][0]));
            b = (BYTE)(atoi(&par[9][0]));
            BMP_triangle((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, x3, y3, r, g, b);
            BMP_triangle((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, x3, y3, r, g, b);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else if((strcmp(&par[0][0], "line") == 0))
    {
        if(dat_bmp_server.ffile > 0)
        {
            x1 = (U32)(atoi(&par[1][0]));
            y1 = (U32)(atoi(&par[2][0]));
            x2 = (U32)(atoi(&par[3][0]));
            y2 = (U32)(atoi(&par[4][0]));
            r = (BYTE)(atoi(&par[5][0]));
            g = (BYTE)(atoi(&par[6][0]));
            b = (BYTE)(atoi(&par[7][0]));
            BMP_Line((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, r, g, b);
            BMP_Line((FILE*)dat_bmp_server.ffile, x1, y1, x2, y2, r, g, b);
            sizeout = sprintf((char*)(dumpo), "Ok\r\n");
        }
        else
           sizeout = sprintf((char*)(dumpo), "file not opened\r\n");
    }
    else
        sizeout = sprintf((char*)(dumpo), "Command Error !\r\n");
    //
    //
    return sizeout;
}





/*************************************************************************/

int BMP_SERVER_Start(U32 port)
{
    
    TCPServerInit(port, &dat_bmp_server.serverfd);
    if(dat_bmp_server.serverfd < 0)
    {
        dat_bmp_server.work = 0;
        dat_bmp_server.stoped = 1;
        return -1;
    }
    else
    {
        dat_bmp_server.stoped = 0;
        dat_bmp_server.work = 1;
        pthread_mutex_init(&mt,NULL);
        pthread_create( &thread1, NULL, (void*)BMP_SERVER_Listen, 0);
        //
        BMP_REDRAW_WORK = 0xFF;
        pthread_create(&thBMP, NULL, (void*)pthBMP_Redraw, 0);
	printf("Server Started \r\n");

    }
    return 0;
}


int BMP_SERVER_Stop(void)
{
    while(dat_bmp_server.stoped == 0)
    {
        dat_bmp_server.work = 0;
        usleep(1);
    }
    return 0;
}

#define BUFFER_LEN	1024
//процесс прослушивания
void BMP_SERVER_Listen(void)
{
    int lwork = 1;
    DAT_BMP_CLIENT client;
    char clientaddr[ BUFFER_LEN];
    //
    /*dat_bmp_server.ffile = 0;
    //
    dat_bmp_server.ffile = fopen((char*)(&DatMain.fname[0]), "rb");
    if(dat_bmp_server.ffile > 0)
    {
        fclose(dat_bmp_server.ffile);
        BMP_FileAnalise((char*)(&DatMain.fname[0]));
        dat_bmp_server.ffile = fopen((char*)(&DatMain.fname[0]),"r+b");
        printf("-- Ok\r\n");
    }
    else
        printf("-- Error open file\r\n");
    */
    //
    while(lwork > 0)
    {
        pthread_mutex_lock(&mt);
        if(dat_bmp_server.work < 1)
        {
            pthread_mutex_unlock(&mt);
            break;
        }
        if( TCPServerWaitConnection(dat_bmp_server.serverfd, &client.clientfd, clientaddr) < 0)
		{
			printf("Bind port error\n");
            pthread_mutex_unlock(&mt);
			break;
		}
        else
        {
            printf("SClient %u\n",client.clientfd);
            pthread_create( &client.pthread, NULL, (void*)BMP_CLIENT, &client);
        }
        pthread_mutex_unlock(&mt);
        usleep(1000);
    }
    TCPServerClose(dat_bmp_server.serverfd);
}

void BMP_CLIENT(DAT_BMP_CLIENT * client)
{
    DAT_BMP_CLIENT cl;
    memcpy(&cl, client, sizeof(DAT_BMP_CLIENT));
    char  bufi[BUFFER_LEN];
    int     inlen, rs;
    char bufo[BUFFER_LEN];
    int     outlen;
    printf("Client %u\n",cl.clientfd);
    //
    /*
    dat_bmp_server.ffile = fopen((char*)(&DatMain.fname[0]), "rb");
    if(dat_bmp_server.ffile > 0)
    {
        fclose(dat_bmp_server.ffile);
        BMP_FileAnalise((char*)(&DatMain.fname[0]));
        dat_bmp_server.ffile = fopen((char*)(&DatMain.fname[0]),"r+b");
        printf("-- Ok\r\n");
    }
    else
        printf("-- Error open file\r\n");
    */
    //
    while(dat_bmp_server.work)
    {
        pthread_mutex_trylock(&mt);
        memset(&bufi[0],0,BUFFER_LEN);
        memset(&bufo[0],0,BUFFER_LEN);
        //
        inlen = 0;
        do{
            rs = TCPNonBlockRead(cl.clientfd, &bufi[inlen], sizeof(bufi) - inlen);
            if(rs > 0)
                inlen += rs;
            else
                break;
        }
        while(1);
        //
        if(inlen >0) 
        {
            //данные приняты
            if(bufi[0] == _id_bin) //бинарный формат - начинается с нулеовго  значения по адресу 0
            {
                //printf("------ binary -----   inlen %u     sended  %u\r\n", inlen, (*(U32*)(&bufi[_size_in_addr])));
                if(inlen == (*(U32*)(&bufi[_size_in_addr]))) //проверка по размеру
                {
                    //printf("----work binary ----\r\n");
                    outlen = iBMP_CLIENT_bin (&bufi[0], &bufo[0]);
                }
                else
                    outlen = 0;
            }
            else //строчный формат
            {
                //printf("------ string -----\r\n");
                outlen = iBMP_CLIENT_string (&bufi[0], &bufo[0]);
                //printf("outlen %i\r\n",outlen);
            }
        }
        else
        {
            outlen = sizeof(int);
            (*(int*)(&bufo[0]))= 0;
            //outlen = 0;
        }
        if(TCPWrite(cl.clientfd, &bufo[0], outlen) < 0)
        {
            pthread_mutex_unlock(&mt);
            break;
        }
        pthread_mutex_unlock(&mt);
        usleep(1000);
    }
    close(cl.clientfd);
    printf("Close Client %u\n",cl.clientfd);
    pthread_exit(&cl.pthread);
}
