//bmp.c

/*
	11.05.208	- переработка с работой только в буфере
			  и цикличным сохранением в файле
*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <stdint.h>
#include "bmp.h"

#pragma pack(1)

BITMAPFILEHEADER bhfi;
BITMAPINFOHEADER bhfo;
BITMAPINFOHEADERr bhfor;
BMP_PAR bmppar;
uint8_t BMP_DUMP[_BMP_DUMP_SIZE_];
uint8_t BMP_REDRAW_WORK;
uint8_t BMP_FILE;


/*
*************************************
* Анализ параметров BMP файла       *
* и загрузка его в буфер            *
*************************************
*/
int BMP_FileAnalise(char *fname)
{
    FILE * file;
    printf("BMP filename:  %s\n", fname);
    file = fopen(fname,"rb");
    if(file < 0) {
        printf("Open %s  error!\n",fname);
        return -1;
    }
    fseek(file, 0, 0);
    fread(&bhfi, sizeof(BITMAPFILEHEADER), 1, file);
    fseek(file, _bcSize, 0);
    fread(&bhfor.biSize, sizeof(uint32_t), 1, file);    
    fseek(file, _bcSize, 0);
    fread(&bhfor.biSize, bhfor.biSize, 1, file);
    //чтение файла в буфер
    fseek(file,0,0);
    fread(&BMP_DUMP[0],1,bhfi.bfSize,file);
    //закрытие файла
    fclose(file);
    BMP_FILE = 0xFF;
    //анализ заголовка INFO
    printf("Size Struct info %u dec   %x hex\n",bhfor.biSize, bhfor.biSize);
    switch(bhfor.biSize)
    {
        case  12:
            printf("Version \"Core\"\n");
            (*(uint64_t*)(&bhfo.biWidth)) = (uint64_t)(*(uint16_t*)(&bhfor.data[0])); 
            (*(uint64_t*)(&bhfo.biHeight)) = (uint64_t)(*(uint16_t*)(&bhfor.data[2])); 
            (*(uint16_t*)(&bhfo.biBitCount)) = (*(uint16_t*)(&bhfor.data[6])); 
            break;
        case 40:
            printf("Version \"3\"\n");
            (*(uint64_t*)(&bhfo.biWidth)) = (*(uint64_t*)(&bhfor.data[0])); 
            (*(uint64_t*)(&bhfo.biHeight)) = (*(uint64_t*)(&bhfor.data[4])); 
            (*(uint16_t*)(&bhfo.biBitCount)) = (*(uint16_t*)(&bhfor.data[10])); 
            break;
        case 108:
            printf("Version \"4\"\n");
            (*(uint64_t*)(&bhfo.biWidth)) = (*(uint64_t*)(&bhfor.data[0])); 
            (*(uint64_t*)(&bhfo.biHeight)) = (*(uint64_t*)(&bhfor.data[4])); 
            (*(uint16_t*)(&bhfo.biBitCount)) = (*(uint16_t*)(&bhfor.data[10])); 
            break;
        case 124:
            printf("Version \"5\"\n");
            (*(uint64_t*)(&bhfo.biWidth)) = (*(uint64_t*)(&bhfor.data[0])); 
            (*(uint64_t*)(&bhfo.biHeight)) = (*(uint64_t*)(&bhfor.data[4])); 
            (*(uint16_t*)(&bhfo.biBitCount)) = (*(uint16_t*)(&bhfor.data[10])); 
            break;
        default:
            break;
    }
    //
    bmppar.step_x = bhfo.biBitCount / 8;
    printf("--- step x %u\r\n", bmppar.step_x);
    bmppar.step_y =(uint64_t)((bhfo.biWidth) * bmppar.step_x);
    printf("--- step y %u\r\n", bmppar.step_y);
    //
    printf("---------------------------------------------------------------------------------\n");
    printf("Type         : %x\n", bhfi.bfType);
    printf("Size          : %u bytes\n", bhfi.bfSize);
    printf("Start addr: %x hex\n", bhfi.bfOffBits);
    printf("---------------------------------------------------------------------------------\n");
    printf("Widht :  %u\n", bhfo.biWidth);
    printf("Haight : %u\n", bhfo.biHeight);
    printf("Bit in Pixel(color) %u\n", bhfo.biBitCount);
    //
    return 1;
}


int BMP_SetPixel(uint64_t x, uint64_t y, uint8_t R, uint8_t G, uint8_t B)
{
	uint32_t addr;
	uint16_t step_x = bmppar.step_x;
	uint64_t step_y = bmppar.step_y;
	RGBQUAD rgb;
	rgb.rgbRed = R;
	rgb.rgbGreen = G;
	rgb.rgbBlue = B;
	//проверка корректности координат
	if((x > bhfo.biWidth) ||
	(y > bhfo.biHeight))
		return -1;
	//
	addr = (step_x * x) + (step_y * y) + bhfi.bfOffBits;
	memcpy(&BMP_DUMP[addr], &rgb, step_x);
	//
	return 0;
}

int BMP_GetPixel(uint64_t x, uint64_t y, uint8_t * R, uint8_t * G, uint8_t * B)
{
	uint32_t addr;
    	uint16_t step_x = bmppar.step_x;
    	uint64_t step_y = bmppar.step_y;
    	RGBQUAD rgb;
	//проверка корректности координат
    	if((x > bhfo.biWidth) ||
        	(y > bhfo.biHeight))
		return -1;
	addr = (step_x * x) + (step_y * y) + bhfi.bfOffBits;
	memcpy(&rgb, &BMP_DUMP[addr], step_x);
	//
	*R = rgb.rgbRed;
	*G = rgb.rgbGreen;
	*B = rgb.rgbBlue; 
	//
	return 0;
}

__inline int iBMP_SetPixel(uint64_t x, uint64_t y, uint8_t R, uint8_t G, uint8_t B)
{
	uint32_t addr;
	uint16_t step_x = bmppar.step_x;
	uint64_t step_y = bmppar.step_y;
	RGBQUAD rgb;
	rgb.rgbRed = R;
	rgb.rgbGreen = G;
	rgb.rgbBlue = B;
	//проверка корректности координат
	if((x > bhfo.biWidth) ||
	(y > bhfo.biHeight))
		return -1;
	//
	addr = (step_x * x) + (step_y * y) + bhfi.bfOffBits;
	memcpy(&BMP_DUMP[addr], &rgb, step_x);
	//
	return 0;
}

__inline void BMP_hline(uint32_t x, uint32_t y, uint32_t len, uint8_t R, uint8_t G, uint8_t B)
{
	uint32_t addr,addr1;
	uint16_t step_x = bmppar.step_x;
	uint64_t step_y = bmppar.step_y;
	RGBQUAD rgb;
	rgb.rgbRed 	= R;
	rgb.rgbGreen 	= G;
	rgb.rgbBlue 	= B;
	addr1 = (step_y * y) + bhfi.bfOffBits;
	while(len--)
	{
		addr = addr1 + (step_x * x);
        	memcpy(&BMP_DUMP[addr], &rgb, step_x);
		x++;
	}    
}


__inline void BMP_vline(uint32_t x, uint32_t y, uint32_t len, uint8_t R, uint8_t G, uint8_t B)
{
	uint32_t addr,addr1;
	uint16_t step_x = bmppar.step_x;
	uint64_t step_y = bmppar.step_y;
	RGBQUAD rgb;
	rgb.rgbRed = R;
	rgb.rgbGreen = G;
	rgb.rgbBlue = B;
	addr1 = (step_x * x) + bhfi.bfOffBits;
	while(len--)
	{
		addr = addr1 + (step_y * y);
		memcpy(&BMP_DUMP[addr], &rgb, step_x);
		y++;
	}

}

/*******
 *  line *
 *******/
 static void do_line(int x1, int y1, int x2, int y2, uint8_t R, uint8_t G, uint8_t B)
{
	int dx = x2-x1;
	int dy = y2-y1;
	int i1, i2;
	int x, y;
	int dd;
	/* worker macro */
	#define DO_LINE(pri_sign, pri_c, pri_cond, sec_sign, sec_c, sec_cond)     \
	{                                                                         \
		if (d##pri_c == 0) {                                                   \
		iBMP_SetPixel(x1, y1, R, G, B);                     \
		return;                                                             \
	}                                                                      \
									     \
	i1 = 2 * d##sec_c;                                                     \
	dd = i1 - (sec_sign (pri_sign d##pri_c));                              \
	i2 = dd - (sec_sign (pri_sign d##pri_c));                              \
									     \
	x = x1;                                                                \
	y = y1;                                                                \
									     \
	while (pri_c pri_cond pri_c##2) {                                      \
		iBMP_SetPixel(x, y, R, G, B);                                                 \
									     \
		if (dd sec_cond 0) {                                                \
			sec_c sec_sign##= 1;                                             \
			dd += i2;                                                        \
		}                                                                   \
		else                                                                \
			dd += i1;                                                        \
									     \
		pri_c pri_sign##= 1;                                                \
		}                                                                      \
	}

	if (dx >= 0) {
		if (dy >= 0) {
			if (dx >= dy) {
				/* (x1 <= x2) && (y1 <= y2) && (dx >= dy) */
				DO_LINE(+, x, <=, +, y, >=);
			}
			else {
				/* (x1 <= x2) && (y1 <= y2) && (dx < dy) */
				DO_LINE(+, y, <=, +, x, >=);
			}
		}
		else {
			if (dx >= -dy) {
				/* (x1 <= x2) && (y1 > y2) && (dx >= dy) */
				DO_LINE(+, x, <=, -, y, <=);
			}
			else {
				/* (x1 <= x2) && (y1 > y2) && (dx < dy) */
				DO_LINE(-, y, >=, +, x, >=);
			}
		}
   	}
   	else {
		if (dy >= 0) {
			if (-dx >= dy) {
				/* (x1 > x2) && (y1 <= y2) && (dx >= dy) */
				DO_LINE(-, x, >=, +, y, >=);
			}
			else {
				/* (x1 > x2) && (y1 <= y2) && (dx < dy) */
				DO_LINE(+, y, <=, -, x, <=);
			}
		}
		else {
			if (-dx >= -dy) {
				/* (x1 > x2) && (y1 > y2) && (dx >= dy) */
				DO_LINE(-, x, >=, -, y, <=);
			}
			else {
				/* (x1 > x2) && (y1 > y2) && (dx < dy) */
				DO_LINE(-, y, >=, -, x, <=);
			}
		}
   	}
}

 
 
void BMP_Line(uint32_t  x1, uint32_t y1, uint32_t  x2, uint32_t y2, uint8_t R, uint8_t G, uint8_t B)
{
     int sx, sy, dx, dy, t;
   
      sx = x1;
      sy = y1;
      dx = x2;
      dy = y2;

      if (sx > dx) {
        t = sx;
        sx = dx;
        dx = t;
      }

      if (sy > dy) {
        t = sy;
        sy = dy;
        dy = t;
      }


    if (x1 == x2) {
      if (y2>y1)
            BMP_vline(x1, y1, y2-y1, R, G, B);
      else
            BMP_vline(x1, y2, y1-y2, R, G, B);
      return;
   }

   if (y1 == y2) {
      if (x2>x1)
         BMP_hline(x1, y1, x2-x1, R, G, B);
      else
        BMP_hline(x2, y1, x1-x2, R, G, B);
      return;
   }

   do_line(x1, y1, x2, y2, R, G, B);
}


/********
 * circle *
 ********/
static void do_circle(int x, int y, int radius, uint8_t R, uint8_t G, uint8_t B)
{
   int cx = 0;
   int cy = radius;
   int df = 1 - radius; 
   int d_e = 3;
   int d_se = -2 * radius + 5;

   do {
       iBMP_SetPixel(x+cx, y+cy, R, G, B);

      if (cx) 
        iBMP_SetPixel(x-cx, y+cy, R, G, B);

      if (cy) 
        iBMP_SetPixel(x+cx, y-cy, R, G, B);

      if ((cx) && (cy)) 
        iBMP_SetPixel(x-cx, y-cy, R, G, B);

      if (cx != cy) {
        iBMP_SetPixel(x+cy, y+cx, R, G, B);

	 if (cx) 
        iBMP_SetPixel(x+cy, y-cx, R, G, B);

	 if (cy) 
        iBMP_SetPixel(x-cy, y+cx, R, G, B);

	 if (cx && cy) 
        iBMP_SetPixel(x-cy, y-cx, R, G, B);
      }

      if (df < 0)  {
        df += d_e;
        d_e += 2;
        d_se += 2;
      }
      else { 
        df += d_se;
        d_e += 2;
        d_se += 4;
        cy--;
      } 

      cx++; 

   } while (cx <= cy);
}

void BMP_Circle(uint32_t x, uint32_t y, uint32_t radius, uint8_t R, uint8_t G, uint8_t B)
{
	do_circle(x, y, radius, R, G, B);
}

void BMP_Circlefill(uint32_t x, uint32_t y, uint32_t radius, uint8_t R, uint8_t G, uint8_t B)
{
   int cx = 0;
   int cy = radius;
   int df = 1 - radius; 
   int d_e = 3;
   int d_se = -2 * radius + 5;
/*
   int sx, sy, dx, dy;
    sx = x-radius-1;
    sy = y-radius-1;
    dx = x+radius+1;
    dy = y+radius+1;
*/
    do
    {
        BMP_hline(x-cy, y-cx, cy*2, R, G, B);
        if (cx)
            BMP_hline(x-cy, y+cx, cy*2, R, G, B);
    	if (df < 0)
        {
            df += d_e;
            d_e += 2;
            d_se += 2;
        }
        else 
        { 
            if (cx != cy)
            {
                BMP_hline(x-cx, y-cy, cx*2, R, G, B);
                if (cy)
                      BMP_hline(x-cx, y+cy, cx*2, R, G, B);
            }

            df += d_se;
            d_e += 2;
            d_se += 4;
            cy--;
        } 
        cx++; 
    } while (cx <= cy);
}

void BMP_Circlefill_fuzzy(uint32_t x, uint32_t y, uint32_t radius, uint8_t R, uint8_t G, uint8_t B, int f)
{
	int cx = 0;
	int cy = radius;
	int df = 1 - radius; 
	int d_e = 3;
	int d_se = -2 * radius + 5;
	//
	do
	{
		BMP_hline(x-cy-f, y-cx, (cy+f)*2, R, G, B);
		if (cx)
			BMP_hline(x-cy-f, y+cx, (cy+f)*2, R, G, B);
		if (df < 0)
		{
			df += d_e;
			d_e += 2;
			d_se += 2;
		}
	else 
	{ 
		if (cx != cy)
		{
			BMP_hline(x-cx-f, y-cy, (cx+f)*2, R, G, B);
			if (cy)
				BMP_hline(x-cx-f, y+cy, (cx +f)*2, R, G, B);
		}
		df += d_se;
		d_e += 2;
		d_se += 4;
		cy--;
	} 
	if(f==1)
		f = -1;
	else
		f = 1;        
	cx++; 
	} while (cx <= cy);
}



/****************
 * RECTANGLE *
 ****************/

void BMP_rect(uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2, uint8_t R, uint8_t G, uint8_t B)
{
    int w, h, t;

    
    if (x1>x2)
    {
        t = x1;
        x1 = x2;
        x2 = t;
    }
    if (y1>y2)
    {
        t = y1;
        y1 = y2;
        y2 = t;
    }

    w = x2 - x1 + 1;
    h = y2 - y1 + 1;
    
   //gfx_hline(x1, y1, w, color);
   BMP_hline(x1, y1, w, R, G, B);
   if (h>1)
	//gfx_hline(x1, y1 + h -1, w, color); 
    BMP_hline(x1, y1+h-1, w, R, G, B);
   //gfx_vline(x1, y1, h, color);
   BMP_vline(x1, y1, h, R, G, B);
   if (w>1)
	//gfx_vline(x1 + w -1, y1, h, color);
    BMP_vline(x1+w-1, y1, h, R, G, B);
}

void BMP_rect_filled(uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2, uint8_t R, uint8_t G, uint8_t B)
{
    int w, h, t;
//    if (drop_next_frame)
//	return;

    if (x1>x2)
    {
        t = x1;
        x1 = x2;
        x2 = t;
    }
    if (y1>y2)
    {
        t = y1;
        y1 = y2;
        y2 = t;
    }

    w = x2 - x1 + 1;
    h = y2 - y1 + 1;
    
    while(h--)    
    {
        //gfx_hline(x1, y1, w, color);
        BMP_hline(x1, y1, w, R, G, B);
        y1++ ;
    }
}


/**************
 * TRIANGLE *
 **************/
void BMP_triangle(uint32_t x1, uint32_t  y1, uint32_t x2, uint32_t y2, uint32_t x3, uint32_t y3, uint8_t R, uint8_t G, uint8_t B)
{
	BMP_Line(x1, y1, x2, y2, R, G, B);
	BMP_Line(x2, y2, x3, y3, R, G, B);
	BMP_Line(x1, y1, x3, y3, R, G, B);
}



