//bmpfunc.c

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include "bmp.h"

pthread_mutex_t         	BMPmutex;
pthread_t 			thBMP;
/*
Определяем параметры файла BMP
и формируем рабочие параметры

входные параметры - имя шаблонного файла BMP
*/
int dumpBMPinit(char *fname)
{
	int res;
	pthread_mutex_lock(&BMPmutex);
	res = BMP_FileAnalise(fname);
	pthread_mutex_unlock(&BMPmutex);
	return res;
}

/*
запись пикселя в буфер BMP

входные параметры:
x,y	- координыты
R,G,B	- параметры цвета пикселя
*/
int setpixel(uint64_t x, uint64_t y, uint8_t R, uint8_t G, uint8_t B)
{
	int res;
	pthread_mutex_lock(&BMPmutex);
	res = BMP_SetPixel(x, y, R, G, B);
	pthread_mutex_unlock(&BMPmutex);
	return res;
}

/*
формирование линии в буфере BMP

входные параметры:
x1,y1,x2,y2	- координаты конечных точек линии
R,G,B		- параметры цвета линии
*/
int line(uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2, uint8_t R, uint8_t G, uint8_t B)
{
	int res = 0;
	pthread_mutex_lock(&BMPmutex);
	BMP_Line(x1, y1, x2, y2, R, G, B);
	pthread_mutex_unlock(&BMPmutex);
	return res;
}

/*
формирование окружности в буфер BMP

входные параметры:
x,y		- координаты центра окружности
radius		- радиус окружности
R,G,B		- параметры цвета
*/
int circle(uint32_t x, uint32_t y, uint32_t radius, uint8_t R, uint8_t G, uint8_t B)
{
	int res = 0;
	pthread_mutex_lock(&BMPmutex);
	BMP_Circle(x, y, radius, R, G, B);
	pthread_mutex_unlock(&BMPmutex);
	return res;
}

/*
 ****************************
 * поток перезаписи файла *
 ****************************
 */
void bmpredraw(char * nfile)
{
	FILE * file;
	char	filename[256]	= {0};
	strcpy(&filename[0], nfile);
	printf("starting loop redraw.....\r\n");
    	while(1)
	{
		file = fopen(&filename[0],"wb");
    		if(file < 0) {
        		printf("Open %s  error!\n",&filename[0]);
        		goto met1;
    		}
		fseek(file, 0, 0);
            	fwrite(&BMP_DUMP[0], 1, bhfi.bfSize, file);
        	fclose(file);
        	//
		printf("---loop---\n");
        	usleep(_interval_redraw);
	}
met1:	printf("stop loop redraw.....\r\n");
	//
	pthread_exit(0);
}

/*
инициализация BMP буфера из файла BMP,
работа обновления данных ф\в файле BMP
*/

int bmpfilework(char * nfile)
{
	int res;
	pthread_mutex_init(&BMPmutex,NULL);
	res = dumpBMPinit(nfile);
	pthread_create(&thBMP, NULL, (void*)bmpredraw, nfile);
	return res;
}



	
